// File _MATH.h
//////////////////////////////////////////////////////////////////////////////
#ifndef  _MATH__
#define  _MATH__

//////////////////////////////////////////////////////////////////////////////
#define _USE_MATH_DEFINES
//#include <cmath>
#include "_TYPE.h"

//////////////////////////////////////////////////////////////////////////////
#define effNO                 0
#define effLINEAR             1  //  

#define effNOTHING         1000  //  
#define effEXPONENT        1001  // 
#define effSEXPONENT       1002  //  
#define effVSEXPONENT      1003  //   
#define effVVSEXPONENT     1004  //    
#define effVVVSEXPONENT    1005  //    
#define effVVVVSEXPONENT   1006  //  ...
#define effVVVVVSEXPONENT  1007  // ...
#define effVVVVVVSEXPONENT 1008  // ...

#define effPOWER          2001  //  
#define effHPOWER         2002  // 
#define effVHPOWER        2003  //  
#define effVVHPOWER       2004  //   
#define effVVVHPOWER      2005  //    

#define effHERMIT          3000  // 

#define effONOFF          4000  //  

#define interpLINEAR      0
#define interpHERMIT      1
#define interpCAVE        4
#define interpHUMP        5

//#define m_PI 3.14159265358979
//#define m_PI 1 / 3.14159265358979

extern tfloat m_PI;
extern tfloat m_1_PI;

extern tfloat TIME_DELTA; 
//////////////////////////////////////////////////////////////////////////////
#define LmB(v, lim)((v > lim) ?(v = lim) : v)  //  "" 
#define LmS(v, lim)((v < lim) ?(v = lim) : v)  //  ""

inline tfloat fClamp(tfloat f, tfloat l) { if (f < l) return f; else return l; }
inline int Round(tfloat f) { if (f < 0) return f - 0.5; else return f + 0.5; }

//////////////////////////////////////////////////////////////////////////////
inline tfloat lExp0(tfloat t, tfloat d) //   
{
  t = (d - t);
  if (t < 0) return 0; else return t / d;
}

inline tfloat lExp0(tfloat t) { return 1 - t; }

//////////////////////////////////////////////////////////////////////////////
inline tfloat shExp0(tfloat t, tfloat d) //     (   )
{
  t = (d - t);
  if (t < 0)  return 0; else { t = t / d * 0.5; d = t * t;
                               return d * (6 - 4 * t);
                             }
}

inline tfloat shExp0(tfloat t) { t = 0.5 * (1 - t); return t * t * (6 - 4 * t); }

//////////////////////////////////////////////////////////////////////////////
inline tfloat hHExp0(tfloat t, tfloat d) //    (   )
{
  t = (d - t);
  if (t < 0)  return 0; else { t = 0.5 * t / d; d = t * t;
                               t = d * (6 - 4 * t);
                               return t * t; 
                             }
}

inline tfloat hHExp0(tfloat t) { t = 0.5 * (1 - t); t = t * t * (6 - 4 * t); return t * t; }

//////////////////////////////////////////////////////////////////////////////
inline tfloat Hill(tfloat t, tfloat d) //  - ..  
{
  t = 2 * t / d; 
  if (t < 1) return t * t * (3 - 2 * t); 
  else { t = 2 - t; return t * t * (3 - 2 * t); };
}
inline tfloat Hill(tfloat t) { t = 2 * t; if (t < 1) return t * t * (3 - 2 * t); else { t = 2 - t; return t * t * (3 - 2 * t); } }

//////////////////////////////////////////////////////////////////////////////
inline tfloat Hump(tfloat t, tfloat d) //  - ..   
{
  t = t / d + 0.5; 
  if (t < 1) return t * t * (6 - 4 * t) - 1; 
  else { t = 2 - t; return t * t * (6 - 4 * t) - 1; };
}
inline tfloat Hump(tfloat t) { t = t + 0.5; if (t < 1) return t * t * (6 - 4 * t) - 1; else { t = 2 - t; return t * t * (6 - 4 * t) - 1; } }

//////////////////////////////////////////////////////////////////////////////
inline tfloat HillHump(tfloat t, tfloat d) //  - ..  
{
  t /= d; 
  if (t < 0.5)
  { t *= 2;  return t * t * (3 - 2 * t); 
  } else { t = 1.5 - t; return t * t * (6 - 4 * t) - 1; }
}

//////////////////////////////////////////////////////////////////////////////
inline tfloat HExp(tfloat t, tfloat d) // ,  ""   , ..  
{
  t = (d - t);
  if (t < 0)  return 0; else { t = t / d; d = t * t;
                               return d * (3 - 2 * t);
                             }
}

inline tfloat HExp(tfloat t) { t = (1 - t); return t * t * (3 - 2 * t); }

//////////////////////////////////////////////////////////////////////////////
inline tfloat pExp(tfloat t, tfloat ds)  //    ds = 1/10
{
  t = t / ds; return 1 / (1 + t * t * 9);
}

inline tfloat pExp(tfloat t) { return 1 / (1 + t * t * 9); }

//////////////////////////////////////////////////////////////////////////////
inline tfloat rExp(tfloat t, tfloat ds) { return exp(-2 / ds * t); }
inline tfloat rExp(tfloat t) { return exp(-2 * t); }

//////////////////////////////////////////////////////////////////////////////
inline tfloat sHermit(tfloat t) //       -  ,  
{
  t *= 0.5; return t * t * (6 - 4 * t);
}

//////////////////////////////////////////////////////////////////////////////
inline tfloat hHermit(tfloat t) //     
{
  t *= 0.5; t = t * t * (6 - 4 * t); return t * t; 
}

//////////////////////////////////////////////////////////////////////////////
inline tfloat Hermit(tfloat t) //  
{
  return t * t * (3 - 2 * t);
}

//////////////////////////////////////////////////////////////////////////////
inline tfloat Power(tfloat t, tfloat p)
{
  p = t + (1 - t) * (1 - p);
  if (p == 0) return 0; else return t / p;
}

//////////////////////////////////////////////////////////////////////////////
inline tfloat mPower(tfloat t, tfloat p)
{
  if (p == 1) { if (t == 0) return t; else return 1;
              } else if (p == 0) return t;
  t = t / (1 - p);
  if (t >= 1) return 1; else { t = 1 - t; return 1 - t * t; }
}

//////////////////////////////////////////////////////////////////////////////
inline tfloat Lower(tfloat t, tfloat l)
{
  l = 1 - t * l;
  if (l == 0) return 1; else return 1 - (1 - t) / l;
}

//////////////////////////////////////////////////////////////////////////////
inline tfloat Dist(tfloat x1, tfloat y1, tfloat x2, tfloat y2)
{
  x2 = x2 - x1; y2 = y2 - y1;
  return sqrt(x2*x2 + y2*y2);
}

//////////////////////////////////////////////////////////////////////////////
inline tfloat ScalarBlend(tfloat s1, tfloat s2, tfloat b) { return s1 + (s2 - s1) * b; } //   s1  s2  b = 0..1

//////////////////////////////////////////////////////////////////////////////
inline tfloat ShiftZero(tfloat t, tfloat sh) { return sh + t * (1 - sh); } //    0..1 ,  0    sh

//////////////////////////////////////////////////////////////////////////////
extern int RandSeed;
inline int RandomSeed(int seed = -1)   
{
  RandSeed = seed;
  if (RandSeed == -1) RandSeed = timeGetTime() & RAND_MAX; 
  return RandSeed;
}

//////////////////////////////////////////////////////////////////////////////
inline tfloat Random(int *seed = &RandSeed)
{ 
  tfloat res;
  seed[0] *= 16807;
  * ((unsigned int *) &res) = (((unsigned int)seed[0])>>9) | 0x40000000;
  return abs(res-3.0f);  
}

inline tfloat mpRandom(int *seed = &RandSeed)   //  -1  +1
{ 
  tfloat res;
  seed[0] *= 16807;
  * ((unsigned int *) &res) = (((unsigned int)seed[0])>>9) | 0x40000000;
  return (res-3.0f);  
}

inline int Random(int m, int *seed = &RandSeed) { return m * Random(seed) + 0.5; }
inline int Random(int min, int max, int *seed = &RandSeed) { return min + (max - min) * Random(seed)+ 0.5; }
inline int RandomSign(int *seed = &RandSeed) { return 1 - 2 * Random(1, seed); }  // -1 or +1

inline int Random(int min, int max, int exclude_zero, int *seed = &RandSeed)
{ 
  if (exclude_zero) return RandomSign(seed);
  else return min + (max - min) * Random(seed) + 0.5;
}

inline tfloat fRandom(tfloat m, int *seed = &RandSeed) { return m * Random(seed); }
inline tfloat fRandom(tfloat min, tfloat max, int *seed = &RandSeed) { return min + (max - min) * Random(seed); }

//////////////////////////////////////////////////////////////////////////////
inline tfloat getAngle(tfloat x, tfloat y)
{
  tfloat a, fx = fabs(x), fy = fabs(y);  
  if (fx < fy)
  { if (fy < 0.0000001) return 0;
    a = atan(fx / fy);
    if (y > 0) { if (x > 0) a += 3*M_PI/2; else a = 3*M_PI/2 - a; } else if (x < 0) a += M_PI/2; else a = M_PI/2 - a; 
  } else
  { if (fx < 0.0000001) return 0;
    a = atan(fy / fx);
    if (y > 0) { if (x > 0) a = 2*M_PI - a; else a += M_PI; } else if (x < 0) a = M_PI - a; else a = a;  
  }
  return a;
}

//inline tfloat getAngle(tfloat x, tfloat y, tfloat r) { r = acos(x / r); if (y > 0) return 2 * M_PI - r; else return r; };

inline tfloat xRotate(tfloat r, tfloat angle) { return r * cos(angle); }
inline tfloat yRotate(tfloat r, tfloat angle) { return - r * sin(angle); }
inline tfloat xRotate(tfloat cx, tfloat r, tfloat angle) { return cx + r * cos(angle); }
inline tfloat yRotate(tfloat cy, tfloat r, tfloat angle) { return cy - r * sin(angle); }
inline void Rotate(tfloat r, tfloat angle, tfloat &x, tfloat &y) { x = r * cos(angle); y = - r * sin(angle); }
inline void Rotate(tfloat cx, tfloat cy, tfloat r, tfloat angle, tfloat &x, tfloat &y) { x = cx + r * cos(angle); y = cy - r * sin(angle); }

inline void Rotate(tfloat angle, tfloat &x, tfloat &y)
{
  tfloat xx = x, yy = y;
  x = xx * cos(angle) - yy * sin(angle);
  y = xx * sin(angle) + yy * cos(angle);
}

inline void Rotate(tfloat cx, tfloat cy, tfloat angle, tfloat &x, tfloat &y)
{
  tfloat xx = x - cx, yy = y - cy;
  x = xx * cos(angle) - yy * sin(angle) + cx;
  y = xx * sin(angle) + yy * cos(angle) + cy;
}

//////////////////////////////////////////////////////////////////////////////
#endif
