// File _RGB.h
//   DWORD     RGB(a)  
//////////////////////////////////////////////////////////////////////////////
#ifndef  _RGB__
#define  _RGB__

#include "_MATH.h" 

//////////////////////////////////////////////////////////////////////////////
typedef DWORD dwRGB;
typedef dwRGB dwRGB_arr[1024];

extern dwRGB commonRGB;

//////////////////////////////////////////////////////////////////////////////
inline unsigned shiftof(unsigned s)
{
  switch(s)
  { case 1: return 0;
    case 2: return 1;
    case 4: return 2;
    case 8: return 3;
    case 16: return 4;
    case 32: return 5;
    case 64: return 6;
    case 128: return 7;
    case 256: return 8;
    case 512: return 9;
    case 1024: return 10;
    case 2048: return 11;
    case 4096: return 12;
    case 8192: return 13;
    case 16384: return 14;
    case 32768: return 15;
    case 65536: return 16;
    case 131072: return 17;
    case 262144: return 18;
    case 524288: return 19;
    case 1048576: return 20;
    case 2097152: return 21;
    case 4194304: return 22;
    case 8388608: return 23;
    case 16777216: return 24;
    case 33554432: return 25;
    case 67108864: return 26;
    case 134217728: return 27;
    case 268435456: return 28;
    case 536870912: return 29;
    case 1073741824: return 30;
    case 2147483648: return 31;
    default: return 888; //  
  }
}

//////////////////////////////////////////////////////////////////////////////
inline int to255(tfloat t) { return t * 255 + 0.5; }

inline int clip255(int i) { return Clamp(i, 0, 255); } //    |0..255|

inline int close255(int i) { return Clamp(i, 255); } //    |255|

//////////////////////////////////////////////////////////////////////////////
#define dwIdent3(dw)((mB(dw) == mB(dw >> 8)) && (mB(dw >> 8)== mB(dw >> 16)))
#define dwIdent4(dw)((mB(dw) == mB(dw >> 8)) && (mB(dw >> 8)== mB(dw >> 16)) && (mB(dw >> 16) == mB(dw >> 24)))

inline dwRGB RGBis(unsigned r, unsigned g, unsigned b) { return (clip255(b) | (clip255(g) << 8) | (clip255(r) << 16) | 0xff000000); }
                           //      [BGRa]
inline dwRGB RGBAis(unsigned r, unsigned g, unsigned b, unsigned a) { return (clip255(b) | (clip255(g) << 8) | (clip255(r) << 16) | (a << 24)); }
                           // c    - 

inline dwRGB ALPHAset(dwRGB reg, unsigned alpha) { return reg & 0x00ffffff | (alpha << 24); }
inline dwRGB ALPHAis(unsigned a) { return (clip255(a) << 24); } // RGBis(10,35,140) + ALPHAis(127)
inline unsigned RGBalpha(dwRGB r) { return (r >> 24); }                           

#define dwRGBoffs(ptr, offs)((dwRGB *)((unsigned)ptr + offs)) 

#define mB(r)(r & 0x000000ff)
#define mG(r)(r & 0x0000ff00)
#define mR(r)(r & 0x00ff0000)
#define mA(r)(r & 0xff000000)

#define sB(r)(r & 0x000000ff)
#define sG(r)((r & 0x0000ff00) >> 8)
#define sR(r)((r & 0x00ff0000) >> 16)
#define sA(r)((r & 0xff000000) >> 24)

//inline dwRGB maxB(dwRGB r) {if (r & 0xffffff00) return 0x000000ff; else return r;};
#define maxB(r)((r & 0xffffff00) ? 0x000000ff : r)  //   > 255
#define minB(r)((r & 0xffffff00) ? 0 : r)           //   < 0

//inline dwRGB maxG(dwRGB r) {if (r & 0xffff00ff) return 0x0000ff00; else return r;};
#define maxG(r)((r & 0xffff00ff) ? 0x0000ff00 : r)  //   > 255
#define minG(r)((r & 0xffff00ff) ? 0 : r)           //   < 0

//inline dwRGB maxR(dwRGB r) {if (r & 0xff00ffff) return 0x00ff0000; else return r;};
#define maxR(r)((r & 0xff00ffff) ? 0x00ff0000 : r)  //   > 255
#define minR(r)((r & 0xff00ffff) ? 0 : r)           //   < 0

inline dwRGB RGBinv(dwRGB r)
{
  return r ^ 0x00ffffff;
};

//     SSE
/*
inline dwRGB RGBadd(dwRGB r1,dwRGB r2)
{
  return maxB(mB(r1) + mB(r2)) | maxG(mG(r1) + mG(r2)) | maxR(mR(r1) + mR(r2));
}

inline dwRGB RGBsub(dwRGB r1, dwRGB r2)
{
  return minB(mB(r1) - mB(r2)) | minG(mG(r1) - mG(r2)) | minR(mR(r1) - mR(r2));
}/**/

inline dwRGB RGBaddV(dwRGB r,int v)            //    
{
  return maxB(mB(r) + v) | maxG(mG(r) + (v << 8)) | maxR(mR(r) + (v << 16));
}

inline dwRGB RGBsubV(dwRGB r, int v)           //    
{
  return minB(mB(r) - v) | minG(mG(r) - (v << 8)) | minR(mR(r) - (v << 16));
}

/*                                                           
inline dwRGB RGBmultV(dwRGB r, int v)          //    
{
  return maxB(mB(r) * v) | maxG(mG(r) * v) | maxR(mR(r) * v);
}                                                        

inline dwRGB RGBmult(dwRGB r1, dwRGB r2)       //       
{                                              // 0 .. 255   0.0 .. 1.0    RGBprocent 
  return mB((mB(r1) * sB(r2)) >> 8) | mG((mG(r1) * sG(r2)) >> 8) | mR((mR(r1) * sR(r2)) >> 8);
}/**/                                                         
                                                           
                                                           
inline dwRGB RGBdivV(dwRGB r, int v) //    
{
  return (mB(r) / v) | mG(mG(r) / v) | mR(mR(r) / v);
}
                                                           
inline dwRGB RGBshiftL(dwRGB r, int v) //  
{
  return maxB(r << v) | maxG(mG(r) << v) | maxR(mR(r) << v);
}
                                                              
inline dwRGB RGBshiftR(dwRGB r, int v) //  
{
  return mB(mB(r) >> v) | mG(mG(r) >> v) | mR(mR(r) >> v);
}
                                                            
inline dwRGB RGBprocent(dwRGB r, unsigned p)
{
  return mB((mB(r) * p) >> 8) | mG((mG(r) * p) >> 8) | mR((mR(r) * p) >> 8);
}
/*                                                           
inline dwRGB RGBblend(dwRGB r1, dwRGB r2, unsigned p)  //  ,       255
{                                                      //  r1  r2, 255 "" - p 
  return mB((int)mB(r1) + (((int)mB(r2) - (int)mB(r1)) * p >> 8)) | mG((int)mG(r1) + (((int)mG(r2) - (int)mG(r1)) * p >> 8)) | mR((int)mR(r1) + (((int)mR(r2) - (int)mR(r1)) * p >> 8));
}/**/
                                                           
//////////////////////////////////////////////////////////////////////////////
//   SSE
//      MMX/*
inline dwRGB RGBadd(dwRGB r1, dwRGB r2)
{
  _asm { movd xmm0,r1
         movd xmm1,r2
         paddusb xmm0,xmm1
         movd r1,xmm0
       }
  return r1;
}

inline dwRGB RGBsub(dwRGB r1, dwRGB r2)
{
  _asm { movd xmm0,r1
         movd xmm1,r2
         psubusb xmm0,xmm1
         movd r1,xmm0
       }
  return r1;
}

inline dwRGB RGBmax(dwRGB r1, dwRGB r2)
{
  _asm { movd xmm0,r1
         movd xmm1,r2
         pmaxub xmm0,xmm1
         movd r1,xmm0
       }
  return r1;
}

inline dwRGB RGBavr(dwRGB r1, dwRGB r2)
{
  _asm { movd xmm0,r1
         movd xmm1,r2
         pavgb xmm0,xmm1
         movd r1,xmm0
       }
  return r1;
}

inline dwRGB RGBavrsoft(dwRGB r1, dwRGB r2)
{
  _asm { movd xmm0,r1
         movd xmm1,r2
         pavgb xmm0,xmm1
         pavgb xmm0,xmm1
         movd r1,xmm0
       }
  return r1;
}

inline dwRGB RGBmult(dwRGB r1, dwRGB r2)
{
  _asm { pxor xmm7,xmm7
         movd xmm0,r1 
         movd xmm1,r2 
         punpcklbw xmm0,xmm7 
         punpcklbw xmm1,xmm7 
         pmullw xmm0,xmm1 
         psrlw xmm0,8    
         packuswb xmm0,xmm0 
         movd r1,xmm0   
       }
  return r1;
}

inline dwRGB RGBimult(dwRGB r1, dwRGB r2)
{
  r1 ^= 0xffffffff; r2 ^= 0xffffffff; 
  _asm { pxor xmm7,xmm7
         movd xmm0,r1 
         movd xmm1,r2 
         punpcklbw xmm0,xmm7 
         punpcklbw xmm1,xmm7 
         pmullw xmm0,xmm1 
         psrlw xmm0,8    
         packuswb xmm0,xmm0 
         movd r1,xmm0   
       }
  return r1 ^ 0xffffffff;
}

inline dwRGB RGBmultV(dwRGB r, int v)
{
  v |= (v << 8) | (v << 16) | (v << 24);
  _asm { pxor xmm7,xmm7
         movd xmm0,r 
         movd xmm1,v
         punpcklbw xmm0,xmm7 
         punpcklbw xmm1,xmm7 
         pmullw xmm0,xmm1 
         psrlw xmm0,8    
         packuswb xmm0,xmm0 
         movd r,xmm0   
       }
  return r;
}  

inline dwRGB RGBblend(dwRGB r1, dwRGB r2, unsigned p)   
{
  p |= (p << 8) | (p << 16) | (p << 24);
  _asm { movd xmm4,p         //  
         punpcklwd xmm4,xmm4 //   c alpha-
         psrlw xmm4,8        //      
         pxor xmm7,xmm7      //  xmm7
         movd xmm0,r2        //   xmm0    
         movd xmm1,r1        //   xmm1    
         punpcklbw xmm0,xmm7 
         punpcklbw xmm1,xmm7 
         psubw xmm0,xmm1     //   ,  
         psllw xmm1,8        //      xmm1  c
         pmullw xmm0,xmm4    //     alpha- 
         paddw xmm1,xmm0     //    
         psrlw xmm1,8        //        
         packuswb xmm1,xmm1  //     32 
         movd r1,xmm1        //     ( )
       }
  return r1;
}

inline dwRGB RGBalphablend(dwRGB r1, dwRGB r2) //     r2
{
  _asm { movd xmm0,r2
         pshufd xmm4,xmm1,0
         psrld xmm4,24
         packssdw xmm4,xmm4
         pxor xmm7,xmm7     
         movd xmm1,r1       
         punpcklbw xmm0,xmm7 
         punpcklbw xmm1,xmm7 
         psubw xmm0,xmm1    
         psllw xmm1,8       
         pmullw xmm0,xmm4   
         paddw xmm1,xmm0    
         psrlw xmm1,8       
         packuswb xmm1,xmm1 
         movd r1,xmm1    
       }
  return r1;
}

/**/
//////////////////////////////////////////////////////////////////////////////
inline dwRGB RGBfloatis(tfloat r, tfloat g, tfloat b) { return RGBis(r * 255 + 0.5, g * 255 + 0.5, b * 255 + 0.5); } 

//////////////////////////////////////////////////////////////////////////////
inline dwRGB RGBfmultV(dwRGB r, tfloat I) { return RGBis(I * sR(r) + 0.5, I * sG(r) + 0.5, I * sB(r) + 0.5); }

//////////////////////////////////////////////////////////////////////////////
inline dwRGB RGBiaddV(dwRGB r, int I) { return RGBis(sR(r) + I, sG(r) + I, sB(r) + I); }

//////////////////////////////////////////////////////////////////////////////
inline void cSwap(dwRGB* &n, dwRGB* &m) { dwRGB *p; if (m < n) { p = m; m = n; n = p; } }
inline void Swap(dwRGB &n, dwRGB &m) { dwRGB p; p = m; m = n; n = p; }

//////////////////////////////////////////////////////////////////////////////
#endif

